import React, { useEffect, useState, useMemo } from 'react';
import { CustomLine } from './CustomLine';
import { RowInformation } from '../types';

interface RowLabelsProps {
  dataRows: RowInformation;
  rowHeight: number;
  yOffset: number;
  rowLabelColumnWidth: number;
  classTopTextColumns: string;
  marginText: number;
  sortState: string;
  headerState: string;
  configStyle: { [key: string]: string };
  onMouseEnterRowLabels: (event: React.MouseEvent<SVGGElement, MouseEvent>) => void;
  onMouseLeaveRowLabels: () => void;
  handleClickHeaderSorting: (event: React.MouseEvent<SVGGElement, MouseEvent>) => void;
}

export const RowLabels = ({
  dataRows,
  rowHeight,
  yOffset,
  rowLabelColumnWidth,
  classTopTextColumns: classTopTextColums,
  marginText,
  sortState,
  headerState,
  configStyle,
  onMouseEnterRowLabels,
  onMouseLeaveRowLabels,
  handleClickHeaderSorting,
}: RowLabelsProps) => {
  const accumulatedWidthHeaders = [0];
  let sum = 0;
  dataRows.forEach((row, index) => {
    sum += row.width;
    if (index !== dataRows.length - 1) {
      accumulatedWidthHeaders.push(sum);
    }
  });
  const [isHovered, setIsHovered] = useState(false);
  const [hoverRowIndex, setHoverRowIndex] = useState<number | null>(null);
  const [iconComponents, setIconComponents] = useState<{ [key: number]: string | undefined }[]>(Array(dataRows.length).fill({}));
  const [iconColors, setIconColors] = useState<string[]>([]);
  const adjustedFontSize = useMemo(() => {
    return Math.max(0.4, (rowHeight * 0.75) / 20);
  }, [rowHeight]);

  useEffect(() => {
    const iconColorsTemporal: string[] = [];

    const updatedIconComponents = dataRows.map((row, indexRows) => {
      let iconComponent: string | undefined = undefined;
      if (row.header == headerState) {
        switch (sortState) {
          case 'asc':
            iconComponent = 'icon-[ic--baseline-arrow-upward]';
            break;
          case 'desc':
            iconComponent = 'icon-[ic--baseline-arrow-downward]';
            break;
          case 'original':
            iconComponent = 'icon-[ic--baseline-sort]';
            break;
          default:
            iconComponent = undefined;
        }
        iconColorsTemporal.push('hsl(var(--clr-sec--800))');
      } else {
        iconComponent = undefined;
        iconColorsTemporal.push('hsl(var(--clr-sec--500))');
      }

      if (indexRows == hoverRowIndex && headerState != row.header) {
        iconComponent = 'icon-[ic--baseline-sort]';
      }

      return { [indexRows]: iconComponent };
    });

    setIconColors(iconColorsTemporal);
    setIconComponents(updatedIconComponents);
  }, [sortState, headerState, hoverRowIndex]);
  return (
    <>
      <g key={'rowLabelsInformation'} className="rowLabelsInformation">
        {dataRows[0] &&
          dataRows[0].data.map((rowLabel, indexRows) => (
            <g key={indexRows} transform={'translate(0,' + (yOffset + indexRows * rowHeight) + ')'}>
              <g className={'rowsLabel row-' + indexRows} onMouseEnter={onMouseEnterRowLabels} onMouseLeave={onMouseLeaveRowLabels}>
                {dataRows.map((row, index) => (
                  <g key={'gRowTable-' + index} transform={'translate(' + accumulatedWidthHeaders[index] + ',0)'}>
                    {row.data[indexRows] !== undefined &&
                    row.data[indexRows] !== '' &&
                    (typeof row.data[indexRows] !== 'object' || Array.isArray(row.data[indexRows])) ? (
                      <>
                        <rect
                          width={dataRows.length - 1 === index ? row.width + rowHeight : row.width}
                          height={rowHeight}
                          fill={indexRows % 2 === 0 ? 'hsl(var(--clr-sec--50))' : 'hsl(var(--clr-sec--0))'}
                          strokeWidth={0}
                        ></rect>
                        <foreignObject x="0" y="0" width={row.width} height={rowHeight}>
                          <div className="w-full h-full flex justify-center items-start">
                            <span className={`${classTopTextColums}`} style={{ fontSize: `${adjustedFontSize}rem` }}>
                              {row.data[indexRows].toString()}
                            </span>
                          </div>
                        </foreignObject>
                      </>
                    ) : (
                      <rect
                        width={dataRows.length - 1 === index ? row.width + rowHeight : row.width}
                        height={rowHeight}
                        fill={'url(#diagonal-lines)'}
                        strokeWidth={0}
                      ></rect>
                    )}
                  </g>
                ))}
              </g>
              <CustomLine x1={0} x2={rowLabelColumnWidth} y1={0} y2={0} strokeWidth={0.025 * rowHeight} className="fill-secondary-300" />
              {indexRows === dataRows[0].data.length - 1 && (
                <CustomLine
                  x1={0}
                  x2={rowLabelColumnWidth}
                  y1={rowHeight}
                  y2={rowHeight}
                  strokeWidth={0.025 * rowHeight}
                  className="fill-secondary-300"
                />
              )}
            </g>
          ))}

        <g key={'rowInformationHeadersLines'} className={'rowInformationHeadersLines'}>
          <CustomLine
            x1={0}
            x2={rowLabelColumnWidth - rowHeight}
            y1={yOffset - rowHeight}
            y2={yOffset - rowHeight}
            strokeWidth={0.025 * rowHeight}
            className="fill-secondary-300"
          />
          <CustomLine
            x1={rowLabelColumnWidth - rowHeight}
            x2={rowLabelColumnWidth}
            y1={yOffset - rowHeight}
            y2={yOffset + 0 * rowHeight}
            strokeWidth={1}
            className="fill-secondary-300"
          />
        </g>
        <g key={'rowInformationHeaders'} className={'rowInformationHeaders'}>
          {dataRows.map((row, indexRows) => (
            <g key={'headersContent-' + indexRows}>
              <g
                key={'gHeadersRows-' + indexRows}
                className={`headersRows-${row.header} cursor-pointer fill-secondary-200 hover:fill-secondary-300`}
                transform={'translate(' + accumulatedWidthHeaders[indexRows] + ',' + (yOffset - rowHeight) + ')'}
                onClick={(event) => {
                  handleClickHeaderSorting(event);
                }}
                onMouseEnter={(event) => {
                  setHoverRowIndex(indexRows);
                  setIsHovered(true);
                }}
                onMouseLeave={(event) => {
                  setIsHovered(sortState !== 'original');
                  setHoverRowIndex(null);
                }}
              >
                <rect
                  width={dataRows.length - 1 === indexRows ? row.width + rowHeight : row.width}
                  height={rowHeight}
                  className="fill-secondary-200"
                  opacity={1.0}
                  strokeWidth={0}
                ></rect>
                <foreignObject x="0" y="0" width={row.width} height={rowHeight}>
                  <div className="w-full h-full flex justify-left">
                    <span className={`${classTopTextColums}`} style={{ fontSize: `${adjustedFontSize}rem` }}>
                      {row.header !== undefined && (typeof row.header !== 'object' || Array.isArray(row.header))
                        ? (row.header as any).toString()
                        : ' '}
                    </span>
                  </div>
                </foreignObject>
                {iconComponents[indexRows] && isHovered && (
                  <svg
                    xmlns="http://www.w3.org/2000/svg"
                    width={2 * row.width - rowHeight}
                    height={rowHeight}
                    style={{ color: iconColors[indexRows], stroke: 'none' }}
                  >
                    {iconComponents[indexRows][indexRows]}
                  </svg>
                )}
              </g>
              <CustomLine
                key={'row-verticalLines-' + indexRows}
                x1={accumulatedWidthHeaders[indexRows]}
                x2={accumulatedWidthHeaders[indexRows]}
                y1={yOffset - rowHeight}
                y2={yOffset + dataRows[0].data.length * rowHeight}
                strokeWidth={0.025 * rowHeight}
                className="fill-secondary-200"
              />
            </g>
          ))}
        </g>
      </g>
    </>
  );
};
